from date_iterator import DateIterator, DateBinner
from persistent_ordered_dict import (
	PersistentOrderedDict, ProgressTracker, DuplicateKeyException, 
	SharedProgressTracker
)
from safe import (
	safe_min, safe_max, safe_lte, safe_lt, safe_gte, safe_gt
)
from tsv import UnicodeTsvReader, UnicodeTsvWriter
from file_utils import ls, file_empty
from managed_process import ManagedProcess
from selenium_crawler import SeleniumCrawler, uses_selenium
from string_alignment import (
	StringAligner, string_distance, string_align, 
	string_align_masks, string_align_path
)
from grouper import chunk, group, flatten
import patterns
from logging import trace
from io import out
from vectorize import Vectorizer



#!/usr/bin/env python

'''
This module enables crawling reddit.com to obtain the html representing
users' profiles.  It is executable and takes one arguments.  To crawl
users, run `./crawl_users.py users`.

The very first time you run this file, you may need to create a tracker,
which is a file that the crawler uses to keep track of which users
have already been crawled.  To do that, run `./crawl_users.py setup`.

The tracker is built from a plain text list of 
active users, which should be listed, one user per line, inside the file 
`active-users.txt`.  This file should be stored in DATA_DIR 
(which is a path specified in the project settings in `SETTINGS.py`).
'''


import time
from pyvirtualdisplay import Display
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait 
from selenium.webdriver.support.select import Select
from selenium.common.exceptions import (
	TimeoutException, NoSuchElementException, WebDriverException, 
	StaleElementReferenceException
)


def uses_selenium(f):
	'''
	This is a decorator, which is intended to be used to decorate methods
	of the Crawler class (or subclasses), which contain a selenium.webdriver
	instance as the attribute `driver`.

	Ensures that the webdriver is closed after a the decorated function 
	completes, or in case an exception occurs within it.
	'''

	def wrapped(self, *args, **kwargs):

		# If needed, make and start a display
		if self.use_display:
			self.display = Display(visible=0, size=(1024, 768))
			self.display.start()

		# make a selenium webdriver, and a wait
		self.driver = webdriver.Firefox()
		self._wait = WebDriverWait(self.driver, self.WAIT_TIME)

		# Run the function in a try block
		try:
			return_val = f(self, *args, **kwargs)

		# Quit driver and stop the display in case of exception...
		except:
			self.driver.quit()
			if self.use_display:
				self.display.stop()
			raise

		# ...Or after function completes
		self.driver.quit()
		if self.use_display:
			self.display.stop()

		# Return the original return value
		return return_val

	return wrapped


class SeleniumCrawler(object):

	REQUEST_DELAY = 1.2
	WAIT_TIME = 30

	# These are added to the crawler's namespace for convevenievce
	# This way they don't need to be imported from Selenium
	Select = Select
	TimeoutException = TimeoutException
	NoSuchElementException = NoSuchElementException
	WebDriverException = WebDriverException
	StaleElementReferenceException = StaleElementReferenceException


	def __init__(self, use_display=True):
		self.use_display = use_display


	def delay(self):
		'''
		Ensures that between any two calls to delay(), at least 
		self.REQUEST_DELAY seconds will have elapsed.
		'''
		if hasattr(self, '_timer'):
			elapsed = time.time() - self._timer
			remaining = self.REQUEST_DELAY - elapsed
			time.sleep(max(0, remaining))

		self._timer = time.time()


	def quit_webdriver(self):
		self.driver.quit()


	def get_page_source(self):
		return self.driver.page_source.encode('utf8')


	def wait(self, f, args=(), kwargs={}):
		return self._wait.until(lambda driver: f(*args, **kwargs))


class UnexpectedPageException(Exception):
	pass



if __name__ == '__main__':
	
	if USE_PYVIRTUALDISPLAY:
		display = Display(visible=0, size=(1024, 768))
		display.start()

	# Run the appropriate function.
	try:
		dispatch()

	# Afterwards, or if there is an exception, stop the display if needed.
	except:
		if USE_PYVIRTUALDISPLAY:
			display.stop()
		raise
	else:
		if USE_PYVIRTUALDISPLAY:
			display.stop()



import multiprocessing


#class ParallelFileProcessor(object):
#
#	def __init__(
#		self,
#		target,
#		in_path,
#		out_path=None,
#		chunksize=100,
#		args=[],
#		kwargs={}
#	):
#		self.file_chunker = FileChunker(fpath, chunksize)
#		if out_path is not None:
#			self.out_file = open(out_path, 'w')
#		else:
#			self.out_file = None
#		
#
#
#class FileChunker(object):
#	def __init__(self, fpath, chunksize):
#		self.fpath = fpath
#		self.f = open(fpath)
#		self.chunksize = chunksize
#
#	def __iter__(self):
#		return self
#
#	def next(self):
#
#		# Accumulate <chunksize> lines into one chunk
#		chunk = []
#		for i in range(self.chunksize):
#
#			try:
#				chunk.append(self.f.next())
#
#			# If we reach the end of the file, return the partial chunk
#			except StopIteration:
#				if len(chunk) > 0:
#					return chunk
#
#				# But if the chunk is empty, then stop iteration
#				else:
#					raise
#
#		return chunk




import t4k

def string_distance(s1, s2):
	sa = StringAligner()
	return sa.string_distance(s1, s2)

def string_align(s1, s2):
	sa = StringAligner()
	distance, path = sa.string_alignment(s1, s2)
	return distance, path

def string_align_masks(s1, s2):
	sa = StringAligner()
	mask1, mask2 = sa.get_string_alignment_masks(s1, s2)
	return 

def string_align_path(s1, s2):
	sa = StringAligner()
	return sa.get_string_alignment_path(s1, s2)


class StringAligner(object):

	UP = 1
	LEFT = 2
	DIAG = 3

	DELETE = 'd'
	INSERT = 'i'
	MATCH = 'm'
	SUBSTITUTE = 's'

	def string_alignment(self, name1, name2):
		distance = [
			[None for j in range(len(name2)+1)] for i in range(len(name1)+1)
		]
		path = [
			[None for j in range(len(name2)+1)] for i in range(len(name1)+1)
		]

		for i in range(len(name1) + 1):
			for j in range(len(name2) + 1):

				# The word boundary matches
				if i == 0 and j == 0:
					distance[i][j] = 0
					path[i][j] = self.DIAG
					continue

				# match character i and j
				match_dist = None
				if i > 0 and j > 0:
					if name1[i-1] == name2[j-1]:
						match_dist = distance[i-1][j-1]

					# substitute character i by char j
					else:
						match_dist = distance[i-1][j-1] + 2

				# Delete one char from name1
				del_dist = None
				if i > 0:
					del_dist = distance[i-1][j] + 1

				# Insert one char from name2
				insert_dist = None
				if j > 0:
					insert_dist = distance[i][j-1] + 1

				distance[i][j] = t4k.safe_min(
					match_dist, del_dist, insert_dist)

				if t4k.safe_lte(match_dist, del_dist):
					if t4k.safe_lte(match_dist, insert_dist):
						path[i][j] = self.DIAG
					else:
						path[i][j] = self.LEFT

				elif t4k.safe_lte(del_dist, insert_dist):
					path[i][j] = self.UP

				else:
					path[i][j] = self.LEFT

		return distance, path



	def string_distance(self, name1, name2):
		distance, path = self.string_alignment(name1, name2)
		return distance[-1][-1]



	def get_string_alignment_masks(self, name1, name2):
		distance, path = self.string_alignment(name1, name2)

		alignment1 = [None for p in range(len(name1))]
		alignment2 = [None for q in range(len(name2))]
		i = len(name1)
		j = len(name2)
		while i > 0 or j > 0:
			if path[i][j] == self.UP:
				i -= 1
				alignment1[i] = 0
				continue

			if path[i][j] == self.LEFT:
				j -= 1
				alignment2[j] = 0
				continue

			if path[i][j] == self.DIAG:
				i -= 1
				j -= 1
				if distance[i][j] == distance[i+1][j+1]:
					alignment1[i] = 1
					alignment2[j] = 1
				else:
					alignment1[i] = 0
					alignment2[j] = 0

		return alignment1, alignment2


	def get_string_alignment_path(self, name1, name2):
		distance, path = self.string_alignment(name1, name2)

		linear_path = []
		i = len(name1)
		j = len(name2)
		while i > 0 or j > 0:
			if path[i][j] == self.UP:
				i -= 1
				linear_path.append(self.DELETE)
				continue

			if path[i][j] == self.LEFT:
				j -= 1
				linear_path.append(self.INSERT)
				continue

			if path[i][j] == self.DIAG:
				i -= 1
				j -= 1
				if distance[i][j] == distance[i+1][j+1]:
					linear_path.append(self.MATCH)
				else:
					linear_path.append(self.SUBSTITUTE)

		linear_path.reverse()
		return linear_path


	def display_string_alignment(self, name1, name2):
		alignment1, alignment2 = self.get_string_alignment_masks(
			name1, name2
		)

		i = 0
		j = 0
		display1 = []
		display2 = []
		display = []
		while i < len(name1) or j < len(name2):

			if i >= len(name1):
				display1.append('-')
				display2.append(name2[j])
				display.append(' ')
				j += 1
				continue

			if j >= len(name2):
				display1.append(name1[i])
				display2.append('-')
				display.append(' ')
				i += 1
				continue

			if alignment1[i]:
				if alignment2[j]:
					display1.append(name1[i])
					display2.append(name2[j])
					display.append('|')
					i += 1
					j += 1

				else:
					display1.append('-')
					display2.append(name2[j])
					display.append(' ')
					j += 1

			else:
				if alignment2[j]:
					display1.append(name1[i])
					display2.append('-')
					display.append(' ')
					i += 1

				else:
					display1.append(name1[i])
					display2.append(name2[j])
					display.append(' ')
					i += 1
					j += 1

		return '\n'.join([
			''.join(display1),
			''.join(display),
			''.join(display2)
		])




#!/usr/bin/env python

import psutil
import time
from multiprocessing import Process, Pipe

DONE = True
NOT_DONE = False
SUICIDE = -1

class ManagedProcess(object):
	'''
	Runs a target function in a subprocess and monitors it.
	The subprocess must call back through a pipe periodically.  If the
	manager doesn't hear from the subprocess after a specified time delay,
	it assumes the subprocess is hung, and so kills the process, and 
	then restarts it by calling the target function again.  
	
	When the target function completes successfully, it should call back
	to the manager and signal that it is done.
	'''

	def __init__(self, target, args=(), kwargs={}, timeout=60):

		# Register args
		self.target = target
		self.args = args
		self.kwargs = kwargs
		self.timeout = timeout


	def _start_managed_process(self):

		# Make a pipe for communication with the subprocess
		self.connection, child_connection = Pipe()
		self.kwargs['pipe'] = child_connection

		# Call target as a separate process
		self.managed_process = Process(
			target=self.target,
			args=self.args,
			kwargs=self.kwargs
		)
		self.managed_process.start()


	def start(self):
		manager_process = Process(target=self._start)
		manager_process.start()

	def _start(self):
		'''
		call <func> inside a subprocess, passing it <args> and <kwargs>.  
		Expect that subprocess to communicate via a pipe every 
		<timeout> seconds.  If no call back is received after 
		<timeout> seconds, kill the subprocess, and restart it
		by calling func again with the original arguments.

		The target function must have an argument called pipe, which
		will be passed one end of a multiprocessing.Pipe().  It doesn't
		matter where that argument occurs in the function signature,
		but any arguments after 'pipe' should be filled using <kwargs>.
		'''

		# Start the subprocess
		self._start_managed_process()

		# Monitor it.  Restart if it crashes or hangs.
		status = NOT_DONE
		while status == NOT_DONE:

			# Listen for a call from the subprocess
			responding = self.connection.poll(self.timeout)

			# Update status if we get a response
			if responding:
				response = self.connection.recv()
				if response == DONE:
					status = DONE

			# If no response, or suicide, restart the process
			if not responding or response == SUICIDE:

				# Give suicidal processes time to die
				if response == SUICIDE:
					time.sleep(3)

				# Try to forcibly kill the process
				try:
					kill_proc_tree(self.managed_process.pid)

				# If the process was already dead, that's OK
				except psutil.NoSuchProcess:
					pass

				# Restart the process
				print 'Manager: respawning process...'
				self._start_managed_process()

		print 'Manager: process finished!'


def kill_proc_tree(pid):    
	'''
	Kill the process having id <pid> and kill all of it's children.
	'''

	# get the process
	parent = psutil.Process(pid)

	# get the children recursively
	children = parent.children(recursive=True)

	# kill all children
	for child in children:
		child.kill()
	psutil.wait_procs(children, timeout=5)

	# now kill the parent
	parent.kill()
	parent.wait(5)






import re
import subprocess
import os


def file_empty(file_path):
	return os.stat(file_path).st_size == 0


class ls(object):

	'''
		Iterator for listing all files in a directory.
		- Lists files found under path, unless <files> is False
		- Can also list directories, if <dirs> is True
		- Only files (and directories) matching <whitelist> 
			will be returned, unless they also match <blacklist>.
		- By default, doesn't descend into subdirectories, unless
			<recurse> is True, lists files in subdirs too.
		- If <recurse> is True, only directories matching <whitelist>
			and not matching <blacklist> will be followed
		- Paths to files are returned relative to <path>, unless <absolute>
			is true, then absolute paths are returned
		- Files are listed in natural sort ordering (i.e. respecting
			numerical order), but if <natural_sort> is False alphabetical
			order is used instead.
		- Hidden files (starting with '.') are ommitted, but if <list_all>
			is true, then they will be included.  In recursive mode, this
			is necessary in orderd to follow hidden  folders.
	'''

	def __init__(
		self,
		path,
		files=True,
		dirs=False,
		whitelist='.*',
		blacklist='^$',
		absolute=False,
		recurse=False,
		list_all=False,
		natural_sort=True,
	):
		self.path = path
		self.files = files
		self.dirs = dirs
		self.whitelist = re.compile(whitelist)
		self.blacklist = re.compile(blacklist)
		self.absolute = absolute
		self.recurse = recurse
		self.list_all = list_all
		self.natural_sort = natural_sort


	def filter_back_pointers(self, dirs):
		'''
			Exclude '.' and '..' from <dirs>.
		'''
		return filter(
			lambda i: not (i.split('/')[-1]=='.' or i.split('/')[-1]=='..'),
			dirs
		)


	def __iter__(self):
		self.yield_files, self.yield_dirs = self._ls(self.path)
		self.visit_dirs = list(self.yield_dirs)
		return self


	def next(self):
		'''
			yield the next item, and absolutize the path if necessary
		'''
		next_item = self.get_next()
		if self.absolute:
			return os.path.abspath(next_item)
		return next_item


	def get_next(self):

		while True:

			# Try popping off the next file, if we're yielding files
			if self.files:
				try:
					return self.yield_files.pop()
				except IndexError:
					pass

			# Try popping off the next dir, if we're yielding dirs
			if self.dirs:
				try:
					return self.yield_dirs.pop()
				except IndexError:
					pass

			# Try looking in the next dir
			if self.recurse:
				try:
					next_dir = self.visit_dirs.pop()
				except IndexError:
					raise StopIteration
				else:
					self.yield_files, self.yield_dirs = self._ls(next_dir)
					self.visit_dirs.extend(self.yield_dirs)

			# If not recursive, then we're done
			else:
				raise StopIteration


	def _ls(self, path, filter_back_pointers=True):

		# Ask OS to list the files and directories in path
		if self.list_all:
			list_command = ['ls', '-a', path]
		else:
			list_command = ['ls', path]
		ls = subprocess.Popen(list_command, stdout=subprocess.PIPE)


		# Sort the files and directories
		if self.natural_sort:
			sort_command = ['sort', '-n', '-r']
		else:
			sort_command = ['sort', '-r']
		items = subprocess.check_output(sort_command, stdin=ls.stdout)
		ls.wait()
		items = [os.path.join(path, i) for i in items.split()]

		# Get rid of '.' and '..', if necessary
		if self.list_all and filter_back_pointers:
			items = self.filter_back_pointers(items)

		# Go through the list separate files from directories
		files = []
		dirs = []
		for item in items:

			# Skip if it doesn't match whitelist, or does match blacklist
			if (
				not self.whitelist.search(item) 
				or self.blacklist.search(item)
			):
				continue

			# Append files to the list of files
			if os.path.isfile(item):
				files.append(item)

			# Append directories to the list of dirs
			else:
				dirs.append(item)

		return files, dirs




#!/usr/bin/env python
from setuptools import setup

setup(
	name='t4k',
	version='0.3.1',
	description='my personal swiss army knife',
	author='edward newell',
	author_email='edward.newell@gmail.com',
	packages=['t4k'],
	license='MIT',
	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Intended Audience :: Science/Research',
		'Intended Audience :: Information Technology',
		'Programming Language :: Python :: 2.7',
		'Natural Language :: English',
		'Topic :: Utilities',
	],
	keywords='trix are for kids'
)




import math

def flatten(iterable, recurse=False, depth=0):
	'''
	flattens an iterable of iterables into a simple list.  E.g.
	turns a list of list of elements into a simple list of elements.

	<recurse> will attempt to flatten the elements within the inner 
	iterable too, and will continue so long as the elements found are 
	iterable.  However, recurse will treat strings as atomic, even though
	they are iterable.
	'''

	flat_list = []
	for element in iterable:

		# If the element is not iterable, just add it to the flat list
		try:
			iter(element)
		except:
			flat_list.append(element)
			continue
		else:

			# If the element is a string, treat it as not iterable
			if isinstance(element, basestring):
				flat_list.append(element)

			# If the element is iterable and <recurse> is True, recurse
			elif recurse:
				flat_list.extend(flatten(element, recurse))

			# If the element is iterable but <recurse> is False, extend
			else:
				flat_list.extend(element)

	return flat_list



def group(iterable, num_chunks):
	'''
	Breaks <iterable> into <num_chunks> chunks (lists) of approximately 
	equal size (chunks will differ by at most one item when <num_chunks> 
	doesn't evenly divide len(<iterable>)).
	'''

	# Coerce the sequence into a list
	iterable = list(iterable)

	# Figure out the chunksize
	num_items = len(iterable)
	chunk_size = num_items / float(num_chunks)

	# Allocate approximately chunksize elements to each chunk, rounded off.
	chunks = []
	for i in range(num_chunks):
		start_index = int(math.ceil(i * chunk_size))
		end_index = int(math.ceil((i+1) * chunk_size))
		chunks.append(iterable[start_index:end_index])

	return chunks


def chunk(iterable, chunk_size):
	'''
	Returns a list of lists of items from iterable, where the internal 
	lists are each approximately of size chunk_size (except the last one,
	which is smaller if chunk_size doesn't evenly divide len(iterable)
	'''

	iterator = iter(iterable)
	chunks = []
	this_chunk = []
	still_has_items = True
	while still_has_items:
		try:
			this_chunk.append(iterator.next())
		except StopIteration:
			still_has_items = False

		if len(this_chunk) == chunk_size:
			chunks.append(this_chunk)
			this_chunk = []

		elif not still_has_items and len(this_chunk) > 0:
			chunks.append(this_chunk)

	return chunks






'''
The persistent ordered dict is a data struture that can be interacted with
like a python dict, and which is easy to keep synced with an on-disk copy

This allows you to easily persist data between non-concurrent runs of a
program.  It's useful for keeping track of progress in long jobs.
'''


import atexit
import signal
import time
import sys
import multiprocessing
import subprocess
import json
import os
import re
import copy
from file_utils import ls


class DuplicateKeyException(Exception):
	pass


class PersistentOrderedDict(object):

	LINES_PER_FILE = 1000
	ESCAPE_TAB_PATTERN = re.compile('\t')
	UNESCAPE_TAB_PATTERN = re.compile(r'(?P<prefix>^|[^\\])\\t')
	ESCAPE_SLASH_PATTERN = re.compile(r'\\')
	UNESCAPE_SLASH_PATTERN = re.compile(r'\\\\')


	def __init__(self, path):
		self.path = path

		# if the path doesn't exist, make it
		if not os.path.exists(path):
			os.makedirs(path)

		# if the path exists but points to a file, raise
		elif os.path.isfile(path):
			raise IOError(
				'The path given to PersistentOrderedDict should correspond '
				'to a folder.  A file was found instead: %s' % path
			)

		# read in all data (if any)
		self.read()

		# Hold is off by default -> updates are immediately written to file
		self._hold = False

		# Keep track of files whose contents don't match values in memory
		self.dirty_files = set()

		# Always be sure to synchronize before the script exits
		atexit.register(self.unhold)


	def hold(self):
		'''
		temporarily prevent writing updates to file
		'''
		self._hold = True

	
	def unhold(self):
		'''
		Resume writing updates to file, synchronize any dirty files.
		'''
		self._hold = False
		self.sync()


	def keys(self):
		return copy.copy(self.key_order)


	def values(self):
		return copy.deepcopy([self.data[k] for k in self.key_order])


	def read(self):

		self.key_order = []
		self.index_lookup = {}
		self.data = {}

		i=0
		for fname in ls(self.path):

			# ensure that files are in expected order,
			# that none are missing, and that no lines are missing.
			try:
				assert(fname == os.path.join(self.path, '%d.json' % i))
			except AssertionError:
				print fname
				print os.path.join(self.path, '%d.json' % i)
				raise
			if i > 0:
				prev_file_path = os.path.join(self.path, '%d.json' % (i-1))
				num_lines_prev_file = len(
					open(prev_file_path, 'r').readlines()
				)
				if num_lines_prev_file != self.LINES_PER_FILE:
					raise ValueError(
						"PersistentOrderedDict: "
						"A file on disk appears to be corrupted, because "
						"it's missing lines: %s " % prev_file_path
					)

			i += 1

			for entry in open(os.path.join(fname)):

				# skip blank lines (there's always one at end of file)
				if entry=='':
					continue

				key, json_record = entry.split('\t', 1)
				key = self.UNESCAPE_TAB_PATTERN.sub('\g<prefix>\t', key)
				key = self.UNESCAPE_SLASH_PATTERN.sub(r'\\', key)
				key = key.decode('utf8')
				
				# remove the newline of the end of json_record, and read it
				record = json.loads(json_record[:-1])
				self.data[key] = record
				self.key_order.append(key)
				self.index_lookup[key] = len(self.key_order)-1


	def mark_dirty(self, key):

		key = self.ensure_unicode(key)
		index = self.index_lookup[key]
		file_num = index / self.LINES_PER_FILE
		self.dirty_files.add(file_num)


	def sync(self):

		#graceful = GracefulDeath()

		# No synchronization happens when hold is on.  This reduces I/O
		# when many values need to be updated
		if self._hold:
			return

		for file_num in self.dirty_files:

			# Get the dirty file
			fname =  '%d.json' % file_num
			f = open(os.path.join(self.path, fname), 'w')

			# Go through keys mapped to this file and re-write them
			start = file_num * self.LINES_PER_FILE
			stop = start + self.LINES_PER_FILE
			for key in self.key_order[start:stop]:

				record = self.data[key]
				# Escape tabs in key, and encode using utf8
				key = self.escape_key(key)
				f.write('%s\t%s\n' % (key, json.dumps(record)))

		# No more dirty files
		self.dirty_files = set()
		#if graceful.kill_now:
		#	print 'dying gracefully'
		#	sys.exit(0)


	def escape_key(self, key):
		key = key.encode('utf8')
		key = self.ESCAPE_SLASH_PATTERN.sub(r'\\\\', key)
		key = self.ESCAPE_TAB_PATTERN.sub(r'\\t', key)
		return key


	def __iter__(self):
		self.pointer = 0
		return self


	def next(self):
		try:
			key = self.key_order[self.pointer]
		except IndexError:
			raise StopIteration

		val = self.data[key]
		self.pointer += 1

		return key, val


	def __contains__(self, key):
		key = self.ensure_unicode(key)
		return key in self.index_lookup


	def __len__(self):
		return len(self.key_order)


	def __getitem__(self, key):
		key = self.ensure_unicode(key)
		return self.data[key]


	def ensure_unicode(self, key):

		# ensure that the key is string-like
		if not isinstance(key, basestring):
			raise ValueError(
				'Keys must be str or unicode, and will be converted to '
				'unicode type internally.'
			)

		# cast the key into unicode if necessary
		if not isinstance(key, unicode):
			key = key.decode('utf8')

		return key


	def update(self, key):
		'''
		This can be called to ensure that a specific key will be 
		synchronized.  It's helpful if a mutable object is stored at that 
		key, since it could be changed without triggering __setitem__;
		this provides a way to notify PersistentOrderedDict that the value 
		at that key has changed.
		'''
		key = self.ensure_unicode(key)
		self.mark_dirty(key)
		self.sync()


	def set_item(self, key, val):

		key = self.ensure_unicode(key)
		val = copy.deepcopy(val)

		# if there isn't already an entry, we need to allocate a new slot
		if key not in self.data:
			self.key_order.append(key)
			self.index_lookup[key] = len(self.key_order)-1

		# update the value held at <key>
		self.data[key] = val
		self.mark_dirty(key)
		self.sync()


	def __setitem__(self, key, val):

		key = self.ensure_unicode(key)
		val = copy.deepcopy(val)

		# if there isn't already an entry, we need to allocate a new slot
		if key not in self.data:
			self.key_order.append(key)
			self.index_lookup[key] = len(self.key_order)-1

		# update the value held at <key>
		self.data[key] = val
		self.mark_dirty(key)
		self.sync()


	def set(self, key, subkey, value):
		key = self.ensure_unicode(key)
		self[key][subkey] = value
		self.update(key)


	def add(self, key):
		key = self.ensure_unicode(key)
		if key in self:
			raise DuplicateKeyException(
				'PersistentOrderedDict: key "%s" already exists.' % key)
		else:
			self[key] = True

	def convert_to_tracker(self):

		# Rewrite every value to satisfy the form of a progress tracker
		self.hold()
		for key, spec in self:

			# Ensure the value at key is a dict, and add special keys
			if not isinstance(spec, dict):
				self[key] = {'val':spec, '_done':False, '_tries':0}
			else:
				self[key].update({'_done':False, '_tries':0})

			# Mark the key as updated
			self.update(key)

		self.unhold()


class ProgressTracker(PersistentOrderedDict):

	def check_or_add(self, key):
		'''
		checks if there is an entry for key already marked as done
		(returns True if so).  If no entry exists for key, it makes one
		and provides it with a defualt value of _done:False and _tries:0
		'''
		key = self.ensure_unicode(key)
		if key in self:
			if self[key]['_done']:
				return True
			else:
				return False
		else:
			self[key] = {'_done':False, '_tries':0}
			return False


	def check(self, key):
		key = self.ensure_unicode(key)
		if key in self:
			if self[key]['_done']:
				return True
			else:
				return False
		else:
			return False


	def add(self, key):
		key = self.ensure_unicode(key)
		if key in self:
			raise DuplicateKeyException(
				'ProgressTracker: key "%s" already exists.' % key)
		else:
			self[key] = {'_done':False, '_tries':0}


	def increment_tries(self, key):
		key = self.ensure_unicode(key)
		self[key]['_tries'] += 1
		self.update(key)


	def reset_tries(self, key):
		key = self.ensure_unicode(key)
		self[key]['_tries'] = 0
		self.update(key)


	def mark_done(self, key):
		key = self.ensure_unicode(key)
		self[key]['_done'] = True
		self.update(key)


	def mark_not_done(self, key):
		key = self.ensure_unicode(key)
		self[key]['_done'] = False
		self.update(key)



def requires_lock(lock):
	def decorator(f):
		def f_with_lock(*args, **kwargs):
			lock.acquire()
			return_val = f(*args, **kwargs)
			lock.release()
			return return_val

		return f_with_lock
	return decorator


def requires_tracker_open(f):
	def f_that_requires_tracker_open(self, *args, **kwargs):
		if not self.tracker_open:
			raise CalledClosedTrackerException
		return f(self, *args, **kwargs)

	return f_that_requires_tracker_open



class CalledClosedTrackerException(Exception):
	pass



class SharedProgressTracker(object):
	pass

	CLOSE = 0
	LOCK = multiprocessing.RLock()

	def __init__(self, path):

		self.client_pipe, server_pipe = multiprocessing.Pipe()
		self.lock = multiprocessing.Lock()

		# create a real progress_tracker and a listen loop around it
		client_tracker = multiprocessing.Process(
			target=progress_tracker_serve,
			args=(path, server_pipe)
		)
		client_tracker.start()

		self.tracker_open = True

	@requires_lock(LOCK)
	def close(self):
		self.client_pipe.send(self.CLOSE)

	@requires_tracker_open
	def hold(self):
		self.LOCK.acquire()
		self.client_pipe.send(('hold',))
		return self.client_pipe.recv()

	@requires_tracker_open
	def unhold(self):
		self.client_pipe.send(('unhold',))
		return_val = self.client_pipe.recv()
		self.LOCK.release()
		return return_val

	@requires_tracker_open
	@requires_lock(LOCK)
	def read(self):
		self.client_pipe.send(('read',))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def mark_dirty(self, key):
		self.client_pipe.send(('mark_dirty', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def sync(self):
		self.client_pipe.send(('sync',))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def __contains__(self, key):
		self.client_pipe.send(('__contains__', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def __len__(self):
		self.client_pipe.send(('__len__',))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def __getitem__(self, key):
		self.client_pipe.send(('__getitem__', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def update(self, key):
		self.client_pipe.send(('update', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def __setitem__(self, key, val):
		self.client_pipe.send(('__setitem__', key, val))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def check_or_add(self, key):
		self.client_pipe.send(('check_or_add', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def set(self, key, subkey, val):
		self.client_pipe.send(('set', key, subkey, val))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def check(self, key):
		self.client_pipe.send(('check', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def add(self, key):
		self.client_pipe.send(('add', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def increment_tries(self, key):
		self.client_pipe.send(('increment_tries', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def reset_tries(self, key):
		self.client_pipe.send(('reset_tries', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def mark_done(self, key):
		self.client_pipe.send(('mark_done', key))
		return self.client_pipe.recv()

	@requires_tracker_open
	@requires_lock(LOCK)
	def mark_not_done(self, key):
		self.client_pipe.send(('mark_not_done', key))
		return self.client_pipe.recv()


	
def progress_tracker_serve(path, pipe):
	progress_tracker = ProgressTracker(path)
	is_open = True
	while is_open:
		message = pipe.recv()
		if message == SharedProgressTracker.CLOSE:
			is_open = False
		else:
			attr, args = message[0], message[1:]
			pipe.send(getattr(progress_tracker, attr)(*args))

		

# This was used to ensure that the PersistentOrderedDict got the chance
# to synchronize with the filesystem before the script exited (and likewise
# for the ProgressTracker.  Now the atexit module is being used.
#
#class GracefulDeath(object):
#	'''
#	Catches kill singals and acknowledges kill signals so that other
#	clients can finish critical steps before dying
#
#	The client callable should instantiate a GracefulDeath, and then
#	watch it's kill_now flag to know when it should die
#	'''
#	kill_now = False
#	def __init__(self):
#		signal.signal(signal.SIGINT, self.exit_gracefully)
#		signal.signal(signal.SIGTERM, self.exit_gracefully)
#
#	def exit_gracefully(self,signum, frame):
#		self.kill_now = True
